import torch
"""    Specification:
    
    Adaptively sets the thresholds for each coordinate: ACClip ( τ k , m k ) = min { τ k | m k | + ϵ , 1 } m k , τ α k = β 2 τ α k − 1 + ( 1 − β 2 ) | g k | α , (ACClip) where operations are element-wise, ϵ is a small number for numerical stability, and we set α = 1 by default. The use of a coordinate-wise α -moment (as opposed to the traditional second moment) is motivated by our theory and estimates B = E [ | g α k | ] 1 / α in Assumption 2. Since E [ | g α k | ] 1 / α is increasing in α ≥ 1 , using α = 1 is a conservative bound on B when the true α is unknown.    
"""

def clip_gradient(grad, k=0.5):
    return grad.clamp(-k, k)
def ac_clip_gradient(grad, params, clip_factor):
    param_norm = torch.norm(torch.cat([param.view(-1) for param in params]))
    
    # Compute the maximum allowed gradient norm
    max_grad_norm = clip_factor * param_norm
    
    # Compute the norm of the gradients
    grad_norm = torch.norm(grad.view(-1))
    
    if grad_norm > max_grad_norm:
        # If gradient norm is greater than the maximum allowed norm, clip the gradient
        grad = grad * (max_grad_norm / (grad_norm + 1e-6))
    
    return grad


import torch
from dowg.CoordinateDoWG import CoordinateDoWG

def test_coordinate_dowg():
    model = torch.nn.Linear(10, 1)
    optimizer = CoordinateDoWG(model.parameters(), clip=0.01)
    criterion = torch.nn.MSELoss()

    # Create some random input and target tensors
    x = torch.randn(1, 100)
    y = torch.randn(1, 1000)

    # Perform a forward pass
    output = model(x)

    # Compute the loss
    loss = criterion(output, y)

    # Perform a backward pass
    loss.backward()

    # Perform a step with the optimizer
    
    try:
        optimizer.step()
    except: 
        print("Error in CoordinateDoWG.step()")
    


import torch

from .clip import clip_gradient


class ScalarDoWG(torch.optim.Optimizer):
    def __init__(self, params, eps=1e-4, clip=0.5, *args, **kwargs):
        defaults = dict(epsilon=1-eps, lr=1, clip=clip)
        self.epsilon = 1-eps
        super(ScalarDoWG, self).__init__(params, defaults)

    def step(self, closure=None):
        
        state = self.state
        loss = None
        if closure is not None:
            loss = closure()
    
            if "rt2" not in state:
                state["rt2"] = torch.Tensor([self.epsilon])
            if "vt" not in state:
                state["vt"] = torch.Tensor([0])
    
            grad_sq_norm = torch.Tensor([0])
            curr_d2 = torch.Tensor([0])
    
            with torch.no_grad():
                for idx, group in enumerate(self.param_groups):
                    group_state = state[str(idx)]  # convert idx to a string
                    if "x0" not in group_state:
                        group_state["x0"] = [torch.clone(p) for p in group["params"]]
        
                    grad_sq_norm += torch.stack(
                        [(p.grad**2).sum() for p in group["params"]]
                    ).sum()
                    curr_d2 += torch.stack(
                        [
                            ((p - p0) ** 2).sum()
                            for p, p0 in zip(group["params"], group_state["x0"])
                        ]
                    ).sum()
        
                state["rt2"] = torch.max(state["rt2"], curr_d2)
                state["vt"] += state["rt2"] * grad_sq_norm
                rt2, vt = state["rt2"], state["vt"]
    
            for group in self.param_groups:
                for p in group["params"]:
                    gt_hat = rt2 * clip_gradient(p.grad.data.clone(), group["clip"])
                    denom = torch.sqrt(vt).add_(group["epsilon"])
                    p.data.addcdiv_(gt_hat, denom, value=-1.0)
        return loss